class CaseLogValidator < ActiveModel::Validator
  # Validations methods need to be called 'validate_' to run on model save
  # or form page submission
  include Validations::SetupValidations
  include Validations::HouseholdValidations
  include Validations::PropertyValidations
  include Validations::FinancialValidations
  include Validations::TenancyValidations
  include Validations::DateValidations
  include Validations::LocalAuthorityValidations
  include Validations::SubmissionValidations

  def validate(record)
    validation_methods = public_methods.select { |method| method.starts_with?("validate_") }
    validation_methods.each { |meth| public_send(meth, record) }
  end
end

class CaseLog < ApplicationRecord
  include Validations::SoftValidations

  has_paper_trail

  validates_with CaseLogValidator
  before_validation :recalculate_start_year!, if: :startdate_changed?
  before_validation :process_postcode_changes!, if: :postcode_full_changed?
  before_validation :process_previous_postcode_changes!, if: :ppostcode_full_changed?
  before_validation :reset_invalidated_dependent_fields!
  before_validation :reset_location_fields!, unless: :postcode_known?
  before_validation :reset_previous_location_fields!, unless: :previous_postcode_known?
  before_validation :set_derived_fields!
  before_save :update_status!

  belongs_to :owning_organisation, class_name: "Organisation"
  belongs_to :managing_organisation, class_name: "Organisation"
  belongs_to :created_by, class_name: "User"

  scope :filter_by_organisation, ->(org, _user = nil) { where(owning_organisation: org).or(where(managing_organisation: org)) }
  scope :filter_by_status, ->(status, _user = nil) { where status: }
  scope :filter_by_years, lambda { |years, _user = nil|
    first_year = years.shift
    query = filter_by_year(first_year)
    years.each { |year| query = query.or(filter_by_year(year)) }
    query.all
  }
  scope :filter_by_year, ->(year) { where(startdate: Time.zone.local(year.to_i, 4, 1)...Time.zone.local(year.to_i + 1, 4, 1)) }

  scope :filter_by_user, lambda { |selected_user, user|
                           if !selected_user.include?("all") && user.present?
                             where(created_by: user)
                           end
                         }

  scope :filter_by_id, ->(id) { where(id:) }
  scope :filter_by_tenant_code, ->(tenant_code) { where("tenant_code ILIKE ?", "%#{tenant_code}%") }
  scope :filter_by_propcode, ->(propcode) { where("propcode ILIKE ?", "%#{propcode}%") }
  scope :filter_by_postcode, ->(postcode_full) { where("postcode_full ILIKE ?", "%#{postcode_full.gsub(/\s+/, '')}%") }
  scope :search_by, lambda { |param|
                      filter_by_id(param)
                                .or(filter_by_tenant_code(param))
                                .or(filter_by_propcode(param))
                                .or(filter_by_postcode(param))
                    }

  AUTOGENERATED_FIELDS = %w[id status created_at updated_at discarded_at].freeze
  OPTIONAL_FIELDS = %w[first_time_property_let_as_social_housing tenant_code propcode].freeze
  RENT_TYPE_MAPPING = { 0 => 1, 1 => 2, 2 => 2, 3 => 3, 4 => 3, 5 => 3 }.freeze
  RENT_TYPE_MAPPING_LABELS = { 1 => "Social Rent", 2 => "Affordable Rent", 3 => "Intermediate Rent" }.freeze
  HAS_BENEFITS_OPTIONS = [1, 6, 8, 7].freeze
  STATUS = { "not_started" => 0, "in_progress" => 1, "completed" => 2 }.freeze
  NUM_OF_WEEKS_FROM_PERIOD = { 2 => 26, 3 => 13, 4 => 12, 5 => 50, 6 => 49, 7 => 48, 8 => 47, 9 => 46, 1 => 52 }.freeze
  SUFFIX_FROM_PERIOD = { 2 => "every 2 weeks", 3 => "every 4 weeks", 4 => "every month" }.freeze
  RETIREMENT_AGES = { "M" => 67, "F" => 60, "X" => 67 }.freeze
  enum status: STATUS

  def form
    FormHandler.instance.get_form(form_name) || FormHandler.instance.forms.first.second
  end

  def collection_start_year
    return @start_year if @start_year
    return unless startdate

    window_end_date = Time.zone.local(startdate.year, 4, 1)
    @start_year = startdate < window_end_date ? startdate.year - 1 : startdate.year
  end

  def recalculate_start_year!
    @start_year = nil
    collection_start_year
  end

  def form_name
    return unless startdate

    "#{collection_start_year}_#{collection_start_year + 1}"
  end

  def self.editable_fields
    attribute_names - AUTOGENERATED_FIELDS
  end

  def completed?
    status == "completed"
  end

  def not_started?
    status == "not_started"
  end

  def in_progress?
    status == "in_progress"
  end

  def weekly_net_income
    return unless earnings && incfreq

    if net_income_is_weekly?
      earnings
    elsif net_income_is_monthly?
      ((earnings * 12) / 52.0).round(0)
    elsif net_income_is_yearly?
      (earnings / 52.0).round(0)
    end
  end

  def weekly_value(field_value)
    num_of_weeks = NUM_OF_WEEKS_FROM_PERIOD[period]
    return unless field_value && num_of_weeks

    (field_value / 52 * num_of_weeks).round(2)
  end

  def applicable_income_range
    return unless ecstat1

    ALLOWED_INCOME_RANGES[ecstat1]
  end

  def first_time_property_let_as_social_housing?
    first_time_property_let_as_social_housing == 1
  end

  def net_income_refused?
    # 2: Tenant prefers not to say
    net_income_known == 2
  end

  def net_income_is_weekly?
    # 1: Weekly
    !!(incfreq && incfreq == 1)
  end

  def net_income_is_monthly?
    # 2: Monthly
    incfreq == 2
  end

  def net_income_is_yearly?
    # 3: Yearly
    incfreq == 3
  end

  def net_income_soft_validation_triggered?
    net_income_in_soft_min_range? || net_income_in_soft_max_range?
  end

  def given_reasonable_preference?
    # 1: Yes
    reasonpref == 1
  end

  def is_renewal?
    # 1: Yes
    renewal == 1
  end

  def is_general_needs?
    # 1: General Needs
    needstype == 1
  end

  def is_supported_housing?
    # 2: Supported Housing
    needstype == 2
  end

  def has_hbrentshortfall?
    # 1: Yes
    hbrentshortfall == 1
  end

  def postcode_known?
    # 1: Yes
    postcode_known == 1
  end

  def previous_postcode_known?
    # 1: Yes
    previous_postcode_known == 1
  end

  def previous_la_known?
    # 1: Yes
    previous_la_known == 1
  end

  def tshortfall_unknown?
    tshortfall_known == 1
  end

  def is_fixed_term_tenancy?
    [4, 6].include?(tenancy)
  end

  def is_secure_tenancy?
    return unless collection_start_year

    # 1: Secure (including flexible)
    if collection_start_year < 2022
      tenancy == 1
    else
      # 6: Secure - fixed term, 7: Secure - lifetime
      [6, 7].include?(tenancy)
    end
  end

  def is_assured_shorthold_tenancy?
    # 4: Assured Shorthold
    tenancy == 4
  end

  def is_internal_transfer?
    # 1: Internal Transfer
    referral == 1
  end

  def is_relet_to_temp_tenant?
    # 9: Re-let to tenant who occupied same property as temporary accommodation
    rsnvac == 9
  end

  def is_bedsit?
    # 2: Bedsit
    unittype_gn == 2
  end

  def is_shared_housing?
    # 4: Shared flat or maisonette
    # 9: Shared house
    # 10: Shared bungalow
    [4, 9, 10].include?(unittype_gn)
  end

  def has_first_let_vacancy_reason?
    # 15: First let of new-build property
    # 16: First let of conversion, rehabilitation or acquired property
    # 17: First let of leased property
    [15, 16, 17].include?(rsnvac)
  end

  def previous_tenancy_was_temporary?
    # 4: Tied housing or renting with job
    # 6: Supported housing
    # 8: Sheltered accomodation
    # 24: Housed by National Asylum Support Service (prev Home Office)
    # 25: Other
    ![4, 6, 8, 24, 25].include?(prevten)
  end

  def armed_forces_regular?
    # 1: Yes – the person is a current or former regular
    !!(armedforces && armedforces == 1)
  end

  def armed_forces_no?
    # 2: No
    armedforces == 2
  end

  def armed_forces_refused?
    # 3: Person prefers not to say / Refused
    armedforces == 3
  end

  def has_pregnancy?
    # 1: Yes
    !!(preg_occ && preg_occ == 1)
  end

  def pregnancy_refused?
    # 3: Tenant prefers not to say / Refused
    preg_occ == 3
  end

  def is_assessed_homeless?
    # 11: Assessed as homeless (or threatened with homelessness within 56 days) by a local authority and owed a homelessness duty
    homeless == 11
  end

  def is_other_homeless?
    # 7: Other homeless – not found statutorily homeless but considered homeless by landlord
    homeless == 7
  end

  def is_not_homeless?
    # 1: No
    homeless == 1
  end

  def is_london_rent?
    # 2: London Affordable Rent
    # 4: London Living Rent
    rent_type == 2 || rent_type == 4
  end

  def previous_tenancy_was_foster_care?
    # 13: Children's home or foster care
    prevten == 13
  end

  def previous_tenancy_was_refuge?
    # 21: Refuge
    prevten == 21
  end

  def is_reason_permanently_decanted?
    # 1: Permanently decanted from another property owned by this landlord
    reason == 1
  end

  def receives_housing_benefit_only?
    # 1: Housing benefit
    hb == 1
  end

  def receives_housing_benefit_and_universal_credit?
    # 8: Housing benefit and Universal Credit (without housing element)
    hb == 8
  end

  def receives_uc_with_housing_element_excl_housing_benefit?
    # 6: Universal Credit with housing element (excluding housing benefit)
    hb == 6
  end

  def receives_no_benefits?
    # 9: None
    hb == 9
  end

  def receives_universal_credit_but_no_housing_benefit?
    # 7: Universal Credit (without housing element)
    hb == 7
  end

  def ethnic_refused?
    ethnic_group == 17
  end

  def receives_housing_related_benefits?
    receives_housing_benefit_only? || receives_uc_with_housing_element_excl_housing_benefit? ||
      receives_housing_benefit_and_universal_credit?
  end

  def benefits_unknown?
    # 3: Don’t know
    hb == 3
  end

  def local_housing_referral?
    # 3: PRP lettings only - Nominated by local housing authority
    referral == 3
  end

  def is_prevten_la_general_needs?
    # 30: Fixed term Local Authority General Needs tenancy
    # 31: Lifetime Local Authority General Needs tenancy
    [30, 31].any?(prevten)
  end

  def self.to_csv
    CSV.generate(headers: true) do |csv|
      csv << attribute_names

      all.find_each do |record|
        csv << record.attributes.map do |att, val|
          record.form.get_question(att, record)&.label_from_value(val) || val
        end
      end
    end
  end

  def soft_min_for_period
    soft_min = LaRentRange.find_by(start_year: collection_start_year, la:, beds:, lettype:).soft_min
    "#{soft_value_for_period(soft_min)} #{SUFFIX_FROM_PERIOD[period].presence || 'every week'}"
  end

  def soft_max_for_period
    soft_max = LaRentRange.find_by(start_year: collection_start_year, la:, beds:, lettype:).soft_max
    "#{soft_value_for_period(soft_max)} #{SUFFIX_FROM_PERIOD[period].presence || 'every week'}"
  end

  def optional_fields
    OPTIONAL_FIELDS + dynamically_not_required
  end

  (1..8).each do |person_num|
    define_method("retirement_age_for_person_#{person_num}") do
      retirement_age_for_person(person_num)
    end

    define_method("plural_gender_for_person_#{person_num}") do
      plural_gender_for_person(person_num)
    end
  end

  def retirement_age_for_person(person_num)
    gender = public_send("sex#{person_num}".to_sym)
    return unless gender

    RETIREMENT_AGES[gender]
  end

  def plural_gender_for_person(person_num)
    gender = public_send("sex#{person_num}".to_sym)
    return unless gender

    if %w[M X].include?(gender)
      "male and non-binary people"
    elsif gender == "F"
      "females"
    end
  end

private

  PIO = Postcodes::IO.new

  def update_status!
    self.status = if all_fields_completed? && errors.empty?
                    "completed"
                  elsif all_fields_nil?
                    "not_started"
                  else
                    "in_progress"
                  end
  end

  def reset_not_routed_questions
    form.invalidated_page_questions(self).each do |question|
      enabled_questions = form.enabled_page_questions(self)
      enabled_question_ids = enabled_questions.map(&:id)
      if %w[radio checkbox].include?(question.type)
        enabled_answer_options = enabled_question_ids.include?(question.id) ? enabled_questions.find { |q| q.id == question.id }.answer_options : {}
        current_answer_option_valid = enabled_answer_options.present? ? enabled_answer_options.key?(public_send(question.id).to_s) : false
        if !current_answer_option_valid && respond_to?(question.id.to_s)
          public_send("#{question.id}=", nil)
        else
          (question.answer_options.keys - enabled_answer_options.keys).map do |invalid_answer_option|
            public_send("#{invalid_answer_option}=", nil) if respond_to?(invalid_answer_option)
          end
        end
      else
        public_send("#{question.id}=", nil) unless enabled_question_ids.include?(question.id)
      end
    end
  end

  def reset_derived_questions
    dependent_questions = { waityear: [{ key: :renewal, value: 0 }],
                            homeless: [{ key: :renewal, value: 0 }],
                            referral: [{ key: :renewal, value: 0 }],
                            underoccupation_benefitcap: [{ key: :renewal, value: 0 }] }

    dependent_questions.each do |dependent, conditions|
      condition_key = conditions.first[:key]
      condition_value = conditions.first[:value]
      if public_send("#{condition_key}_changed?") && condition_value == public_send(condition_key) && !public_send("#{dependent}_changed?")
        self[dependent] = nil
      end
    end
  end

  def reset_invalidated_dependent_fields!
    return unless form

    reset_not_routed_questions
    reset_derived_questions
  end

  def dynamically_not_required
    not_required = []
    not_required << "previous_la_known" if postcode_known?
    not_required << "tshortfall" if tshortfall_unknown?
    not_required << "tenancylength" if tenancylength_optional?

    not_required
  end

  def tenancylength_optional?
    return false unless collection_start_year
    return true if collection_start_year < 2022

    collection_start_year >= 2022 && !is_fixed_term_tenancy?
  end

  def set_derived_fields!
    # TODO: Remove once we support supported housing logs
    self.needstype = 1 unless needstype
    if rsnvac.present?
      self.newprop = has_first_let_vacancy_reason? ? 1 : 2
    end
    self.incref = 1 if net_income_refused?
    self.renttype = RENT_TYPE_MAPPING[rent_type]
    self.lettype = get_lettype
    self.totchild = get_totchild
    self.totelder = get_totelder
    self.totadult = get_totadult
    self.refused = get_refused
    self.ethnic = 17 if ethnic_refused?
    if %i[brent scharge pscharge supcharg].any? { |f| public_send(f).present? }
      self.brent ||= 0
      self.scharge ||= 0
      self.pscharge ||= 0
      self.supcharg ||= 0
      self.tcharge = brent.to_f + scharge.to_f + pscharge.to_f + supcharg.to_f
    end
    if period.present?
      self.wrent = weekly_value(brent) if brent.present?
      self.wscharge = weekly_value(scharge) if scharge.present?
      self.wpschrge = weekly_value(pscharge) if pscharge.present?
      self.wsupchrg = weekly_value(supcharg) if supcharg.present?
      self.wtcharge = weekly_value(tcharge) if tcharge.present?
      if is_supported_housing? && chcharge.present?
        self.wchchrg = weekly_value(chcharge)
      end
    end
    self.has_benefits = get_has_benefits
    self.tshortfall_known = 0 if tshortfall
    self.wtshortfall = if tshortfall && receives_housing_related_benefits?
                         weekly_value(tshortfall)
                       end
    self.nocharge = household_charge&.zero? ? 1 : 0
    self.housingneeds = get_housingneeds
    if is_renewal?
      self.underoccupation_benefitcap = 2 if collection_start_year == 2021
      self.homeless = 2
      self.referral = 0
      self.waityear = 1
      if is_general_needs?
        # fixed term
        self.prevten = 32 if managing_organisation.provider_type == "PRP"
        self.prevten = 30 if managing_organisation.provider_type == "LA"
      end
    end
    (2..8).each do |idx|
      if age_under_16?(idx)
        self["ecstat#{idx}"] = 9
      elsif public_send("ecstat#{idx}") == 9 && age_known?(idx)
        self["ecstat#{idx}"] = nil
      end
    end
  end

  def age_under_16?(person_num)
    public_send("age#{person_num}") && public_send("age#{person_num}") < 16
  end

  def age_known?(person_num)
    !!public_send("age#{person_num}_known")&.zero?
  end

  def process_postcode_changes!
    self.postcode_full = upcase_and_remove_whitespace(postcode_full)
    process_postcode(postcode_full, "postcode_known", "is_la_inferred", "la")
  end

  def process_previous_postcode_changes!
    self.ppostcode_full = upcase_and_remove_whitespace(ppostcode_full)
    process_postcode(ppostcode_full, "previous_postcode_known", "is_previous_la_inferred", "prevloc")
  end

  def process_postcode(postcode, postcode_known_key, la_inferred_key, la_key)
    return if postcode.blank?

    self[postcode_known_key] = 1
    inferred_la = get_inferred_la(postcode)
    self[la_inferred_key] = inferred_la.present?
    self[la_key] = inferred_la if inferred_la.present?
  end

  def reset_location_fields!
    reset_location(is_la_inferred, "la", "is_la_inferred", "postcode_full", 1)
  end

  def reset_previous_location_fields!
    reset_location(is_previous_la_inferred, "prevloc", "is_previous_la_inferred", "ppostcode_full", previous_la_known)
  end

  def reset_location(is_inferred, la_key, is_inferred_key, postcode_key, is_la_known)
    if is_inferred || is_la_known != 1
      self[la_key] = nil
    end
    self[is_inferred_key] = false
    self[postcode_key] = nil
  end

  def get_totelder
    ages = [age1, age2, age3, age4, age5, age6, age7, age8]
    ages.count { |x| !x.nil? && x >= 60 }
  end

  def get_totchild
    relationships = [relat2, relat3, relat4, relat5, relat6, relat7, relat8]
    relationships.count("C")
  end

  def get_totadult
    total = !age1.nil? && age1 >= 16 && age1 < 60 ? 1 : 0
    total + (2..8).count do |i|
      age = public_send("age#{i}")
      relat = public_send("relat#{i}")
      !age.nil? && ((age >= 16 && age < 18 && %w[P X].include?(relat)) || age >= 18 && age < 60)
    end
  end

  def get_refused
    return 1 if age_refused? || sex_refused? || relat_refused? || ecstat_refused?

    0
  end

  def get_inferred_la(postcode)
    # Avoid network calls when postcode is invalid
    return unless postcode.match(Validations::PropertyValidations::POSTCODE_REGEXP)

    postcode_lookup = nil
    begin
      # URI encoding only supports ASCII characters
      ascii_postcode = postcode.encode("ASCII", "UTF-8", invalid: :replace, undef: :replace, replace: "")
      Timeout.timeout(5) { postcode_lookup = PIO.lookup(ascii_postcode) }
    rescue Timeout::Error
      Rails.logger.warn("Postcodes.io lookup timed out")
    end
    if postcode_lookup && postcode_lookup.info.present?
      postcode_lookup.codes["admin_district"]
    end
  end

  def get_has_benefits
    HAS_BENEFITS_OPTIONS.include?(hb) ? 1 : 0
  end

  def get_lettype
    return unless renttype.present? && needstype.present? && owning_organisation[:provider_type].present?

    case RENT_TYPE_MAPPING_LABELS[renttype]
    when "Social Rent"
      if is_supported_housing?
        owning_organisation[:provider_type] == "PRP" ? 2 : 4
      elsif is_general_needs?
        owning_organisation[:provider_type] == "PRP" ? 1 : 3
      end
    when "Affordable Rent"
      if is_supported_housing?
        owning_organisation[:provider_type] == "PRP" ? 6 : 8
      elsif is_general_needs?
        owning_organisation[:provider_type] == "PRP" ? 5 : 7
      end
    when "Intermediate Rent"
      if is_supported_housing?
        owning_organisation[:provider_type] == "PRP" ? 10 : 12
      elsif is_general_needs?
        owning_organisation[:provider_type] == "PRP" ? 9 : 11
      end
    end
  end

  def get_housingneeds
    return 1 if has_housingneeds?
    return 2 if no_housingneeds?
    return 3 if unknown_housingneeds?
  end

  def has_housingneeds?
    if [housingneeds_a, housingneeds_b, housingneeds_c, housingneeds_f].any?(1)
      1
    end
  end

  def no_housingneeds?
    if housingneeds_g == 1
      1
    end
  end

  def unknown_housingneeds?
    if housingneeds_h == 1
      1
    end
  end

  def all_fields_completed?
    subsection_statuses = form.subsections.map { |subsection| subsection.status(self) }.uniq
    subsection_statuses == [:completed]
  end

  def all_fields_nil?
    not_started_statuses = %i[not_started cannot_start_yet]
    subsection_statuses = form.subsections.map { |subsection| subsection.status(self) }.uniq
    subsection_statuses.all? { |status| not_started_statuses.include?(status) }
  end

  def age_refused?
    [age1_known, age2_known, age3_known, age4_known, age5_known, age6_known, age7_known, age8_known].any?(1)
  end

  def sex_refused?
    [sex1, sex2, sex3, sex4, sex5, sex6, sex7, sex8].any?("R")
  end

  def relat_refused?
    [relat2, relat3, relat4, relat5, relat6, relat7, relat8].any?("R")
  end

  def ecstat_refused?
    [ecstat1, ecstat2, ecstat3, ecstat4, ecstat5, ecstat6, ecstat7, ecstat8].any?(10)
  end

  def soft_value_for_period(value)
    num_of_weeks = NUM_OF_WEEKS_FROM_PERIOD[period]
    return "" unless value && num_of_weeks

    (value * 52 / num_of_weeks).round(2)
  end

  def upcase_and_remove_whitespace(string)
    string.present? ? string.upcase.gsub(/\s+/, "") : string
  end
end
